@@ -28,8 +28,6 @@ from urllib.parse import urlparse
 VISOR_DIR = os.path.expanduser('~/.visor')
 VISOR_CONFIG = os.path.join(VISOR_DIR, 'config.json')
 
-# TODO: when running lots of builds in parallel, it seems like duplicate network
-# ids can be created, despite the locks on _create_namespaced_network()
 # TODO: make ctrl+c properly cancel a build
 # TODO: don't take snapshots of very fast commands
 # TODO: partition by team
@@ -38,11 +36,9 @@ VISOR_CONFIG = os.path.join(VISOR_DIR, 'config.json')
 # TODO: SNAPSHOT should be global/local to a machine
 # TODO: --keep-running vms should auto-sleep after 20 mins of inactivity
 # TODO: total time getting printed before error
-# TODO: COPY semantics should match docker exactly
 # TODO: clean up network after all containers in that network are gone
 # TODO: host CPU going crazy even when guest cpu is close to idle
-# TODO: make memory adjustable from inside visorfile
-# TODO: server should be able to specify root dir
+# TODO: server should be able to specify root dir instead of ~/.visorserver
 
 config = {}
 if os.path.exists(VISOR_CONFIG):
@@ -81,7 +77,7 @@ def _stream(resp):
   client = sseclient.SSEClient(resp)
   for event in client.events():
     if event.event == 'log':
-      print(json.loads(event.data))
+      print(json.loads(event.data), end='')
     elif event.event == 'error':
       print(json.loads(event.data))
       exitcode = 1
@@ -229,9 +225,9 @@ def find_context_files(visorfile):
 # t - type
 # v - value
 # p - permissions
-def create_mapped_file(filemap, path):
-  parts = Path(path).parts
-  curpath = ''
+def create_mapped_file(context_dir, filemap, context_relative_path):
+  parts = Path(context_relative_path).parts
+  curpath = context_dir
   for part in parts[:-1]:
     curpath = os.path.join(curpath, part)
     if part in filemap:
@@ -246,52 +242,51 @@ def create_mapped_file(filemap, path):
   filemap[parts[-1]] = {}
   return filemap[parts[-1]]
 
-def chunkify(files):
+def chunkify(context_dir, files):
   chunks = {}
   filemap = {
     't': 'd',
     'v': {},
   }
   for f in files:
-    if os.path.islink(f):
-      mapped_file = create_mapped_file(filemap['v'], f)
+    abs_file = os.path.join(context_dir, f)
+    if os.path.islink(abs_file):
+      mapped_file = create_mapped_file(context_dir, filemap['v'], f)
       mapped_file['t'] = 'l'
-      mapped_file['v'] = os.readlink(f)
+      mapped_file['v'] = os.readlink(abs_file)
       # Symlink does not have its own permissions, permissions of target
       # are used instead
       continue
-    if os.path.isdir(f):
+    if os.path.isdir(abs_file):
       raise Exception(f'Transmitting directories not supported {f}')
-    mapped_file = create_mapped_file(filemap['v'], f)
+    mapped_file = create_mapped_file(context_dir, filemap['v'], f)
     mapped_file['t'] = 'f'
     mapped_file['v'] = []
-    mapped_file['p'] = os.stat(f)[ST_MODE]
-    for chunk in fastcdc(f, 2**11, 2**14, 2**15, hf=blake3.blake3, fat=True):
+    mapped_file['p'] = os.stat(abs_file)[ST_MODE]
+    for chunk in fastcdc(abs_file, 2**11, 2**14, 2**15, hf=blake3.blake3, fat=True):
       mapped_file['v'].append(chunk.hash)
       chunks[chunk.hash] = chunk.data
   return filemap, chunks
 
-def transmit_context(visorfile, archive, server_url=None):
-  server_url = verify_login(server_url)
-  c = config['auths'][server_url]
-  with tempfile.NamedTemporaryFile('w', delete=False) as f:
-    f.write(c['sshKey'])
-    f.flush()
-    with tempfile.NamedTemporaryFile('w', delete=False) as g:
-      files = find_context_files(visorfile)
-      filemap, chunks = chunkify(files)
-      missing_chunks = safe_non_stream_api_call(
-        'post', 'transmitmeta', json={'filemap': filemap}
-      )['missing']
-      safe_non_stream_api_call(
-        'post',
-        'transmit',
-        json={
-          'chunks': {
-            x: base64.b64encode(chunks[x]).decode('ascii') for x in missing_chunks
-          }
+def transmit_context(visorfile, archive):
+  files = find_context_files(visorfile)
+  context_dir = find_context_dir(visorfile)
+  filemap, chunks = chunkify(context_dir, files)
+  missing_chunks = safe_non_stream_api_call(
+    'post', 'transmitmeta', json={'filemap': filemap}
+  )['missing']
+  MAX_CHUNKS_PER_UPLOAD = 1000
+  for i in range(0, len(missing_chunks), MAX_CHUNKS_PER_UPLOAD):
+    safe_non_stream_api_call(
+      'post',
+      'transmit',
+      json={
+        'chunks': {
+          x: base64.b64encode(chunks[x]).decode('ascii') for x in
+              missing_chunks[i:i + MAX_CHUNKS_PER_UPLOAD]
         }
-      )
+      }
+    )
   return filemap
 
 def do_build_base(args):
@@ -327,6 +322,7 @@ def do_build(args):
       'archive': archive,
       'visorfileRelpath': os.path.relpath(visorfile, start=context_dir),
       'keepRunning': args.keep_running,
+      'snapshotOnFailure': args.snapshot_on_failure,
       'buildenvs': buildenvs,
       # TODO: transmitting the filemap once here and once above in the
       # transmitmeta. Could keep something on the server so we only have to
@@ -381,14 +377,35 @@ def do_console(args):
 
 def do_ps(args):
   resp = safe_non_stream_api_call('get', 'ps')
-  for container in resp['containers']:
-    print(container)
+  if args.quiet:
+    for container in resp['containers']:
+      print(container['id'])
+  else:
+    print(f'{"CONTAINER ID":<35}', end='')
+    print(f'{"MACHINE ID":<15}', end='')
+    print(f'{"NETWORK ID":<15}', end='')
+    print(f'{"STATUS":<10}', end='')
+    print(f'{"STARTED":<15}', end='')
+    print('')
+    for container in resp['containers']:
+      print(f'{container["id"]:<35}', end='')
+      print(f'{container["machineId"]:<15}', end='')
+      print(f'{container["networkId"]:<15}', end='')
+      print(f'{container["status"]:<10}', end='')
+      print(f'{container["birth"]:<15}', end='')
+      print('')
 
-def do_stop(args):
-  raise Exception('Not implemented')
+def do_pause(args):
+  resp = safe_non_stream_api_call(
+    'post', 'pause', json={'containerId': args.container_id}
+  )
+  print(resp['containerId'])
 
-def do_start(args):
-  raise Exception('Not implemented')
+def do_unpause(args):
+  resp = safe_non_stream_api_call(
+    'post', 'unpause', json={'containerId': args.container_id}
+  )
+  print(resp['containerId'])
 
 def do_rm(args):
   resp = safe_non_stream_api_call(
@@ -407,16 +424,10 @@ def do_rmi(args):
   )
   print(resp['imageId'])
 
-def do_view(args):
-  resp = safe_non_stream_api_call(
-    'post', 'view', json={'publishableId': args.publishable_id}
-  )
-  print(f'Published guest port {resp["guestPort"]} on localhost:{resp["hostPort"]}')
-
-def do_clean(args):
+def do_wipe(args):
   safe_stream_api_call(
     'post',
-    'clean',
+    'wipe',
     json={
       'includeBases': args.include_bases,
     },
@@ -457,6 +468,7 @@ def main():
 
   build_parser = subparsers.add_parser('build', help='Build a Visorfile')
   build_parser.add_argument('--keep-running', action='store_true')
+  build_parser.add_argument('--snapshot-on-failure', action='store_true')
   build_parser.add_argument('--build-env', dest='buildenvs', action='append', default=[])
   build_parser.add_argument('visorfile', help='The Visorfile to build')
   build_parser.set_defaults(func=do_build)
@@ -477,13 +489,16 @@ def main():
   console_parser.set_defaults(func=do_console)
 
   ps_parser = subparsers.add_parser('ps', help='List running containers')
+  ps_parser.add_argument('--quiet', '-q', action='store_true')
   ps_parser.set_defaults(func=do_ps)
 
-  stop_parser = subparsers.add_parser('stop', help='Stop a running container')
-  stop_parser.set_defaults(func=do_stop)
+  unpause_parser = subparsers.add_parser('unpause', help='Start a running container')
+  unpause_parser.add_argument('container_id', type=str)
+  unpause_parser.set_defaults(func=do_unpause)
 
-  start_parser = subparsers.add_parser('start', help='Start a stopped container')
-  start_parser.set_defaults(func=do_start)
+  pause_parser = subparsers.add_parser('pause', help='Pause a running container')
+  pause_parser.add_argument('container_id', type=str)
+  pause_parser.set_defaults(func=do_pause)
 
   rm_parser = subparsers.add_parser('rm', help='Kill a container')
   rm_parser.add_argument('container_id', type=str)
@@ -497,13 +512,9 @@ def main():
   rmi_parser.add_argument('image_id', type=str)
   rmi_parser.set_defaults(func=do_rmi)
 
-  view_parser = subparsers.add_parser('view', help='View a published id')
-  view_parser.add_argument('publishable_id', type=str)
-  view_parser.set_defaults(func=do_view)
-
-  clean_parser = subparsers.add_parser('clean', help='Wipe all images and vms')
-  clean_parser.add_argument('--include-bases', action='store_true')
-  clean_parser.set_defaults(func=do_clean)
+  wipe_parser = subparsers.add_parser('wipe', help='Wipe all images and vms')
+  wipe_parser.add_argument('--include-bases', action='store_true')
+  wipe_parser.set_defaults(func=do_wipe)
 
   args = main_parser.parse_args()
   args.func(args)
